"""
functions specific to the controller
"""

from ..queries.decorators import wrap_rethink_errors, wrap_connection
from ..queries import RPC, RPP
from .. import r
from ..decorators import deprecated_function
from ..checks import verify
from ..brain_pb2 import Plugin, Port
from .decorators import expect_arg_type, set_plugin_id
from .helpers import _check_common, has_port_conflict
from .interfaces import get_ports_by_ip
from . import DESIRE_ACTIVE, DESIRE_STOP, DESIRE_RESTART
from . import DESIRED_STATE_KEY, ALLOWED_DESIRED_STATES
from . import ADDRESS_KEY, NAME_KEY, SERVICE_KEY, ID_KEY
from .verification import verify_port_map


DEFAULT_LOOKUP_KEY = "Name"


def verify_plugin_contents(plugin):
    """
    extra validation /
    PB2 sees empty lists the same as non-existant lists
    :param plugin:
    :return:
    """
    result = True
    result &= verify_port_map(plugin)
    return result


@deprecated_function(replacement="brain.controller.plugins.find_plugin")
@wrap_rethink_errors
@wrap_connection
def get_plugin_by_name(plugin_name,
                       conn=None):
    """

    :param plugin_name: <str> name of plugin
    :param conn: <rethinkdb.DefaultConnection>
    :return: <list> rethinkdb cursor
    """
    result = RPC.filter({
        "Name": plugin_name
    }).run(conn)
    return result


@wrap_rethink_errors
@wrap_connection
def find_plugin(value,
                key=DEFAULT_LOOKUP_KEY,
                conn=None):
    """
    get's the plugin matching the key and value

    example: find_plugin("plugin1", "ServiceName") => list of 0 or 1 item
    example: find_plugin("plugin1", "Name") => list of 0-to-many items

    :param value:
    :param key: <str> (default "Name")
    :param conn:
    :return:
    """
    # cast to list to hide rethink internals from caller
    result = list(RPC.filter({
        key: value
    }).run(conn))
    return result


@wrap_rethink_errors
@wrap_connection
def get_plugins(conn=None):
    """
    :param conn: <rethinkdb.DefaultConnection>
    :return: <list>
    """
    return list(RPC.run(conn))


def get_names(conn=None):
    """

    :param conn:
    :return: <list> of <str>
    """
    return list(set([x[NAME_KEY] for x in get_plugins(conn=conn)]))


@wrap_rethink_errors
@wrap_connection
def create_plugin(plugin_data,
                  verify_plugin=False,
                  conn=None):
    """
    :param plugin_data: <dict> dict matching Plugin()
    :param verify_plugin: <bool>
    :param conn: <rethinkdb.DefaultConnection>
    :return: <dict> rethinkdb insert response value
    """
    assert isinstance(plugin_data, dict)
    if verify_plugin and not verify(plugin_data, Plugin()):
        raise ValueError("Invalid Plugin entry")
    current = find_plugin(plugin_data[SERVICE_KEY], SERVICE_KEY, conn)
    if not current:
        success = RPC.insert(plugin_data,
                             conflict="update").run(conn)
    else:
        success = {
            "errors": 1,
            "first_error": "Duplicate service name exists {}".format(
                plugin_data[SERVICE_KEY])}
    return success


@deprecated_function(replacement=
                     "brain.controller.interfaces._get_existing_interface")
def _get_existing_interface(existing, port_data):
    """

    :param existing: <list>
    :param port_data:  <dict>
    :return: interface or None
    """
    from .interfaces import _get_existing_interface as _new_gei
    return _new_gei(existing, port_data)


@deprecated_function(replacement="brain.controller.interfaces.create_port")
def create_port(port_data,
                verify_port=False,
                conn=None):
    """

    :param port_data: <dict> dict matching Port()
    :param verify_port: <bool>
    :param conn: <rethinkdb.DefaultConnection>
    :return: <dict> rethinkdb insert response value
    """
    from .interfaces import create_port as new_create_port
    return new_create_port(port_data, verify_port=verify_port, conn=conn)


@deprecated_function(replacement="brain.controller.interfaces.get_interfaces")
def get_interfaces(conn=None):
    """

    :param conn:
    :return:
    """
    from .interfaces import get_interfaces as new_get_interfaces
    return new_get_interfaces(conn=conn)


@expect_arg_type(expected=(dict, ))
@wrap_rethink_errors
@wrap_connection
@set_plugin_id
def update_plugin(plugin_data,
                  verify_plugin=False,
                  conn=None):
    """

    :param plugin_data: <dict> dict matching Plugin()
    :param verify_plugin: <bool>
    :param conn: <rethinkdb.DefaultConnection>
    :return: <dict> rethinkdb update response value
    """

    update_id = plugin_data[ID_KEY]
    original = get(update_id, conn=conn)
    success = RPC.get(update_id).update(plugin_data).run(conn)
    if verify_plugin:
        updated = get(update_id, conn=conn)
        if not verify(updated, Plugin()):
            success = RPC.get(update_id).update(original).run(conn)
            raise ValueError("Invalid Plugin entry {}".format(success))
    return success


@wrap_rethink_errors
@wrap_connection
def get(plugin_id, conn=None):
    """

    :param plugin_id:
    :param conn:
    :return:
    """
    return RPC.get(plugin_id).run(conn)


@wrap_rethink_errors
@wrap_connection
def quick_change_desired_state(plugin_id, desired_state, conn=None):
    """

    :param plugin_id:
    :param desired_state:
    :param conn:
    :return:
    """
    assert desired_state in ALLOWED_DESIRED_STATES
    desired = {DESIRED_STATE_KEY: desired_state}
    return RPC.get(plugin_id).update(desired).run(conn)


def activate(plugin_id, conn=None):
    """

    :param plugin_id:
    :param conn:
    :return:
    """
    return quick_change_desired_state(plugin_id,
                                      DESIRE_ACTIVE,
                                      conn=conn)


def restart(plugin_id, conn=None):
    """

    :param plugin_id:
    :param conn:
    :return:
    """
    return quick_change_desired_state(plugin_id,
                                      DESIRE_RESTART,
                                      conn=conn)


def stop(plugin_id, conn=None):
    """

    :param plugin_id:
    :param conn:
    :return:
    """
    return quick_change_desired_state(plugin_id,
                                      DESIRE_STOP,
                                      conn=conn)



"""
decorator objects for other functions in the query module
"""
from decorator import decorator
from ..connection import rethinkdb as r, connect
from ..connection import validate_get_dbs


WRAP_RETHINK_ERRORS = (r.errors.ReqlOpFailedError,
                       r.errors.ReqlError,
                       r.errors.ReqlDriverError)
STATUS_FIELD = "Status"
START_FIELD = "StartTime"


@decorator
def wrap_job_cursor(func_, *args, **kwargs):
    """
    wraps a filter generator.
    Types should be appropriate before passed to rethinkdb

    somewhat specific to the _jobs_cursor function

    :param func_: <callable function>
    :param args: <must have at least 3 arguments>
    :param kwargs:
    :return: returned value from func_
    """
    assert isinstance(args[0], str)
    assert isinstance(args[1], (str, type(None)))
    assert isinstance(args[2], (str, type(None)))
    if args[2] and not args[1]:
        raise ValueError("Must specify location if using port.")
    return func_(*args, **kwargs)


@decorator
def wrap_rethink_errors(func_, *args, **kwargs):
    """
    Wraps rethinkdb specific errors as builtin/Brain errors

    :param func_: <function> to call
    :param args:  <tuple> positional arguments
    :param kwargs: <dict> keyword arguments
    :return: inherits from the called function
    """
    try:
        return func_(*args, **kwargs)
    except WRAP_RETHINK_ERRORS as reql_err:
        raise ValueError(str(reql_err))

@decorator
def wrap_rethink_generator_errors(func_, *args, **kwargs):
    """

    :param func_: <function> to call
    :param args:  <tuple> positional arguments
    :param kwargs: <dict> keyword arguments
    :return: inherits from the called function
    """
    try:
        for data in func_(*args, **kwargs):
            yield data
    except WRAP_RETHINK_ERRORS as reql_err:
        raise ValueError(str(reql_err))


@decorator
def wrap_connection_reconnect_test(func_, *args, **kwargs):
    """
    if a connection argument is passed, verify connection is good
    if not connected, attempt reconnect
    if reconnect fails, raises the rethink error
    other decorator will translate to standard error

    note: <rethinkdb.DefaultConnection>.is_open() appears to hang
            when the database is pulled out from under conn obj

    should be decorated prior to wrap_connection

    :param func_: <function> to call
    :param args:  <tuple> positional arguments
    :param kwargs: <dict> keyword arguments
    :return:
    :raises: ReqlDriverError or ConnectionRefusedError if reconnect failed.
    """
    conn = args[-1]
    if conn:  # conn object attempted
        try:
            validate_get_dbs(conn)
        except WRAP_RETHINK_ERRORS:
            conn.reconnect()  #throw may occur here
    return func_(*args, **kwargs)


@decorator
@wrap_connection_reconnect_test
def wrap_connection(func_, *args, **kwargs):
    """
    conn (connection) must be the last positional argument
    in all wrapped functions

    :param func_: <function> to call
    :param args:  <tuple> positional arguments
    :param kwargs: <dict> keyword arguments
    :return:
    """
    if not args[-1]:
        new_args = list(args)
        new_args[-1] = connect()
        args = tuple(new_args)
    return func_(*args, **kwargs)



"""

"""
from . import EX_PORTS_KEY, IN_PORTS_KEY, PORT_SEPARATOR, \
    ENV_KEY, ENV_SEPARATOR
from . import ALLOWED_PROTOCOLS


def verify_port_map(plugin):
    """
    extra validation /
    PB2 sees empty lists the same as non-existant lists
    :param plugin:
    :return:
    """
    result = True
    ex_ports = plugin.get(EX_PORTS_KEY)
    result &= isinstance(ex_ports, list) and verify_ports(ex_ports)
    in_ports = plugin.get(IN_PORTS_KEY)
    result &= isinstance(in_ports, list) and verify_ports(in_ports)
    env_list = plugin.get(ENV_KEY)
    result &= isinstance(env_list, list) and verify_environment(env_list)
    if result:
        result &= len(ex_ports) == len(in_ports)
    return result


def verify_ports(port_list):
    """
    ports should be of the format
    <int:port_number>/<str:protocol>
    where protocol is either tcp or udp

    :param port_list: <list>
    :return: <bool>
    """
    result = True  # empty list is ok
    for item in port_list:
        try:
            (port, protocol) = item.split(PORT_SEPARATOR)
            port_num = int(port)
        except ValueError:
            result = False
            break
        result &= protocol in ALLOWED_PROTOCOLS
        result &= port == str(port_num)
    return result


def verify_environment(env_list):
    """

    :param env_list: <list>
    :return: <bool>
    """
    result = True  # empty list is ok
    for item in env_list:
        try:
            (key, value) = item.split(ENV_SEPARATOR)
        except ValueError:
            result = False
            break
        result &= bool(key)
        result &= bool(value)
    return result



"""
Provides a very thin wrapper to rethinkdb as well as common helper functions
"""
import rethinkdb, rethinkdb as r  # rethinkdb passthrough don't delete

from .connection import connect
from .environment import check_log_env, check_prod_env, check_dev_env, log_env_gte, check_stage_env

from . import binary
from . import controller
from . import queries





"""
assortment of wrapped queries
"""
from ..brain_pb2 import Jobs, Target, Commands
from ..checks import verify
from ..jobs import WAITING, STATES, transition_success
from ..connection import rethinkdb as r
from ..decorators import deprecated_function
from .decorators import wrap_connection, wrap_rethink_errors
from .decorators import START_FIELD, STATUS_FIELD
from . import RPX, RBT, RBJ, RPC, RPP, RBO
from .reads import plugin_exists, get_job_by_id

VALID_STATES = STATES


def waiting_filter(lte_time):
    """
    generates a filter for status==waiting and
    time older than lte_time

    :param lte_time: <float> time()
    :return:
    """
    return ((r.row[STATUS_FIELD] == WAITING) &
            (r.row[START_FIELD] <= lte_time))


@deprecated_function(replacement="brain.controller.plugins.has_port_conflict")
def _check_port_conflict(port_data,
                         existing):
    from ..controller.plugins import has_port_conflict
    return has_port_conflict(port_data, existing)


@wrap_rethink_errors
@wrap_connection
def insert_target(target, verify_target=False,
                  conn=None):
    """

    :param target:  <dict> of Target() format
    :param verify_target: <bool>
    :param conn: <rethinkdb.DefaultConnection>
    :return: <dict> rethinkdb insert response value
    """
    if verify_target and not verify(target, Target()):
        raise ValueError("Invalid Target")
    return RBT.insert([target]).run(conn)


@wrap_rethink_errors
@wrap_connection
def insert_jobs(jobs, verify_jobs=True, conn=None):
    """
    insert_jobs function inserts data into Brain.Jobs table

    jobs must be in Job format

    :param jobs: <list> of Jobs
    :param verify_jobs: <bool>
    :param conn: <rethinkdb.DefaultConnection>
    :return: <dict> rethinkdb insert response value
    """
    assert isinstance(jobs, list)
    if verify_jobs and not verify({"Jobs": jobs}, Jobs()):
        raise ValueError("Invalid Jobs")
    inserted = RBJ.insert(jobs).run(conn)
    return inserted


@wrap_rethink_errors
@wrap_connection
def update_job_status(job_id, status, conn=None):
    """Updates a job to a new status

    :param job_id: <str> the id of the job
    :param status: <str> new status
    :param conn: <connection> a database connection (default: {None})

    :return: <dict> the update dicts for the job and the output
    """
    if status not in VALID_STATES:
        raise ValueError("Invalid status")
    job_update = RBJ.get(job_id).update({STATUS_FIELD: status}).run(conn)
    if job_update["replaced"] == 0 and job_update["unchanged"] == 0:
        raise ValueError("Unknown job_id: {}".format(job_id))
    id_filter = (r.row["OutputJob"]["id"] == job_id)
    output_job_status = {"OutputJob": {"Status": status}}
    output_update = RBO.filter(id_filter).update(output_job_status).run(conn)
    return {str(RBJ): job_update, str(RBO): output_update}


@wrap_rethink_errors
@wrap_connection
def write_output(job_id, content, conn=None):
    """writes output to the output table

    :param job_id: <str> id of the job
    :param content: <str> output to write
    :param conn:
    """
    output_job = get_job_by_id(job_id, conn)
    results = {}
    if output_job is not None:
        entry = {
            "OutputJob": output_job,
            "Content": content
        }
        results = RBO.insert(entry, conflict="replace").run(conn)
    return results


@wrap_rethink_errors
@wrap_connection
def create_plugin(plugin_name, conn=None):
    """
    Creates a new plugin

    :param plugin_name: <str>
    :param conn:
    :return: <dict> rethinkdb response
    """
    results = {}
    if not plugin_exists(plugin_name, conn=conn):
        results = RPX.table_create(plugin_name,
                                   primary_key="CommandName").run(conn)
    return results


@wrap_rethink_errors
@wrap_connection
def destroy_plugin(plugin_name, conn=None):
    """
    Creates a new plugin

    :param plugin_name: <str>
    :param conn:
    :return: <dict> rethinkdb response
    """
    results = {}
    if plugin_exists(plugin_name, conn=conn):
        results = RPX.table_drop(plugin_name).run(conn)
    return results


@wrap_rethink_errors
@wrap_connection
def advertise_plugin_commands(plugin_name, commands,
                              verify_commands=False,
                              conn=None):
    """

    :param plugin_name:
    :param commands:
    :param verify_commands:
    :param conn:
    :return:
    """
    assert isinstance(commands, list)
    if verify_commands and not verify({"Commands": commands},
                                      Commands()):
        raise ValueError("Invalid Commands")
    plugin = RPX.table(plugin_name)
    success = plugin.insert(commands,
                            conflict="update").run(conn)
    return success


@deprecated_function(replacement="brain.controller.plugins.create_plugin")
def create_plugin_controller(plugin_data,
                             verify_commands=False,
                             conn=None):
    from ..controller.plugins import create_plugin
    return create_plugin(plugin_data,
                         verify_plugin=verify_commands,
                         conn=conn)


@deprecated_function(replacement="brain.controller.plugins.create_port")
def create_port_controller(port_data,
                           verify_port=False,
                           conn=None):
    from ..controller.plugins import create_port
    return create_port(port_data,
                       verify_port=verify_port,
                       conn=conn)


@deprecated_function(replacement="brain.controller.plugins.update_plugin")
def update_plugin_controller(plugin_data,
                             verify_plugin=False,
                             conn=None):
    from ..controller.plugins import update_plugin
    return update_plugin(plugin_data,
                         verify_plugin=verify_plugin,
                         conn=conn)


@wrap_rethink_errors
@wrap_connection
def transition_waiting(start_time, conn=None):
    """

    :param start_time: <float> from time.time()
    :return: <dict>
    :raises may raise reqlerror, wrapped into ValueError by decorator
    """
    wait_filter = waiting_filter(start_time)
    status_change = {STATUS_FIELD: transition_success(WAITING)}
    return RBJ.filter(wait_filter).update(status_change).run(conn)



"""
code to interact with the (misnomer)
r.Controller.Ports table
"""

from .. import r
from ..queries.decorators import wrap_connection, wrap_rethink_errors
from ..queries import RPP
from ..checks import verify
from ..brain_pb2 import Port
from .helpers import has_port_conflict, _check_common
from .decorators import expect_arg_type
from . import ADDRESS_KEY, UDP_KEY, TCP_KEY, CONFLICT_ACTION, ID_KEY


@wrap_rethink_errors
@wrap_connection
def get_ports_by_ip(ip_address,
                    conn=None):
    """

    :param ip_address: <str> name of interface
    :param conn: <rethinkdb.DefaultConnection>
    :return: <list> rethinkdb cursor
    """
    if ip_address == "":
        result = RPP.filter({
            "Address": ip_address
        }).run(conn)
    else:
        result = RPP.filter(
            (r.row[ADDRESS_KEY] == ip_address) |
            (r.row[ADDRESS_KEY] == "")
        ).run(conn)
    return result


@wrap_rethink_errors
@wrap_connection
def get_interfaces(conn=None):
    """

    :param conn:
    :return:
    """
    res = RPP.pluck([ADDRESS_KEY]).run(conn)
    return list(set([x[ADDRESS_KEY] for x in res]))  # unique list


def _get_existing_interface(existing, port_data):
    """

    :param existing: <list>
    :param port_data:  <dict>
    :return: interface or None
    """
    interface_existing = None
    for interface in existing:
        if interface[ADDRESS_KEY] == port_data[ADDRESS_KEY]:
            interface_existing = interface
    return interface_existing


@expect_arg_type(expected=(dict, ))
@wrap_rethink_errors
@wrap_connection
def create_port(port_data,
                verify_port=False,
                conn=None):
    """

    :param port_data: <dict> dict matching Port()
    :param verify_port: <bool>
    :param conn: <rethinkdb.DefaultConnection>
    :return: <dict> rethinkdb insert response value
    """
    if verify_port and not verify(port_data, Port()):
        raise ValueError("Invalid Port entry")
    existing = list(get_ports_by_ip(port_data[ADDRESS_KEY], conn=conn))
    conflicts = has_port_conflict(port_data, existing)
    if conflicts:
        return conflicts
    interface_existing = _get_existing_interface(existing, port_data)
    if not interface_existing:
        success = RPP.insert(
            port_data,
            conflict=CONFLICT_ACTION
        ).run(conn)
    else:
        combined_tcp = port_data[TCP_KEY] + interface_existing[TCP_KEY]
        interface_existing[TCP_KEY] = list(set(combined_tcp))
        combined_udp = port_data[UDP_KEY] + interface_existing[UDP_KEY]
        interface_existing[UDP_KEY] = list(set(combined_udp))
        success = RPP.get(interface_existing[ID_KEY]).update({
            TCP_KEY: interface_existing[TCP_KEY],
            UDP_KEY: interface_existing[UDP_KEY]
        }).run(conn)
    return success



"""
controls Job related changes
"""
from decorator import decorator
from .decorators import verify_jobs_args_is_tuple, verify_jobs_args_length
from .decorators import COMMAND_FIELD, INPUT_FIELD, OPTIONAL_FIELD
from .decorators import VALUE_FIELD
from .brain_pb2 import Job, Jobs
from .checks import verify

BEGIN = ""
INVALID = "Invalid"
VALID = "Valid"
READY = "Ready"
STOP = "Stopped"
PENDING = "Pending"
DONE = "Done"
ERROR = "Error"
WAITING = "Waiting"
ACTIVE = "Active"

SUCCESS = "success"
FAILURE = "failure"
TRANSITION = "transition"


class JobsError(Exception):
    """
    Base Jobs exception
    """


class InvalidStateTransition(JobsError):
    """
    Simple exception to identify invalid state transition

    State Transition governed by The Brain Documentation
    """
    pass


class InvalidState(JobsError):
    """
    Simple exception to identify invalid state transition

    State Transition governed by The Brain Documentation
    """
    pass


STATES = {BEGIN: {SUCCESS: READY,
                  FAILURE: ERROR,
                  TRANSITION: frozenset([VALID,
                                         READY,
                                         INVALID,
                                         ERROR,
                                         STOP])},
          INVALID: {SUCCESS: VALID,
                    FAILURE: INVALID,
                    TRANSITION: frozenset([VALID,
                                           INVALID])},
          VALID: {SUCCESS: READY,
                  FAILURE: INVALID,
                  TRANSITION: frozenset([READY,
                                         WAITING,
                                         INVALID])},
          READY: {SUCCESS: PENDING,
                  FAILURE: ERROR,
                  TRANSITION: frozenset([STOP,
                                         PENDING,
                                         ERROR])},
          PENDING: {SUCCESS: DONE,
                    FAILURE: ERROR,
                    TRANSITION: frozenset([STOP,
                                           ERROR,
                                           ACTIVE,
                                           DONE])},
          DONE: {SUCCESS: DONE,
                 FAILURE: ERROR,
                 TRANSITION: frozenset([DONE,
                                        ERROR])},
          WAITING: {SUCCESS: READY,
                    FAILURE: ERROR,
                    TRANSITION: frozenset([READY,
                                           ERROR,
                                           STOP])},
          ACTIVE: {SUCCESS: DONE,
                   FAILURE: ERROR,
                   TRANSITION: frozenset([ERROR,
                                          DONE])},
          STOP: {SUCCESS: STOP,
                 FAILURE: ERROR,
                 TRANSITION: frozenset([STOP,
                                        ERROR,
                                        WAITING,
                                        READY])},
          ERROR: {SUCCESS: DONE,
                  FAILURE: ERROR,
                  TRANSITION: frozenset([DONE,
                                         ERROR,
                                         STOP])}}


@decorator
def wrap_good_state(func_, *args, **kwargs):
    """
    Decorator/Wrapper to verify the input is an acceptable state
    prior to calling a function on it

    :param f: <function>  to call
    :param args: <tuple> positional args
    :param kwargs: <dict> keyword args
    :return: wrapped function return
    """
    if not verify_state(args[0]):
        raise InvalidState("{} is not a valid state".format(args[0]))
    return func_(*args, **kwargs)


def verify_state(state):
    """
    Verifies a state is acceptable

    :param state: <str> state name
    :return: <bool>
    """
    return state in STATES


@wrap_good_state
def transition_success(state):
    """
    transition to the on-success state

    :param state: <str> state name
    :return: <str> or None
    """
    return STATES[state][SUCCESS]


@wrap_good_state
def transition_fail(state):
    """
    transition to the on-failure state

    :param state: <str> state name
    :return: <str> or None
    """
    return STATES[state][FAILURE]


@wrap_good_state
def transition(prior_state, next_state):
    """
    Transitions to a non-standard state

    Raises InvalidStateTransition if next_state is not allowed.

    :param prior_state: <str>
    :param next_state: <str>
    :return: <str>
    """
    if next_state not in STATES[prior_state][TRANSITION]:
        acceptable = STATES[prior_state][TRANSITION]
        err = "cannot {}->{} may only {}->{}".format(prior_state,
                                                     next_state,
                                                     prior_state,
                                                     acceptable)
        raise InvalidStateTransition(err)
    return next_state


def verify_job(job):
    """
    verifies a job

    :param job: <dict>
    :return: <bool>
    """
    return verify(job, Job())


def verify_jobs(jobs):
    """
    verifies list of jobs

    :param job: <list>
    :return: <bool>
    """
    return verify(jobs, Jobs())


def _get_args_loop(job, key):
    args_tuple = list()
    for job_args in job[COMMAND_FIELD][key]:
        args_tuple.append(job_args[VALUE_FIELD])
    return args_tuple


def get_args(job):
    """
    This function gets the arguments from a job
    :param job: job dictionary
    :return: input tuple, optional tuple
    """
    return tuple(_get_args_loop(job, INPUT_FIELD)), \
        tuple(_get_args_loop(job, OPTIONAL_FIELD))


def _apply_args_loop(job, args, key):
    for i in range(len(args)):
        job[COMMAND_FIELD][key][i][VALUE_FIELD] = args[i]


@verify_jobs_args_length
@verify_jobs_args_is_tuple
def apply_args(job, inputs, optional_inputs=None):
    """
    This function is error checking before the job gets
    updated.
    :param job: Must be a valid job
    :param inputs: Must be a tuple type
    :param optional_inputs: optional for OptionalInputs
    :return: job
    """
    _apply_args_loop(job, inputs, INPUT_FIELD)
    _apply_args_loop(job, optional_inputs, OPTIONAL_FIELD)
    return job



"""
manipulate the controller database
"""

NAME_KEY = "Name"
ID_KEY = "id"
SERVICE_KEY = "ServiceName"
UDP_KEY = "UDPPorts"
TCP_KEY = "TCPPorts"

CONFLICT_ACTION = "update"

AVAILABLE = "Available"
ACTIVE = "Active"
STOPPED = "Stopped"
RESTARTING = "Restarting"


DESIRE_ACTIVE = "Activate"
DESIRE_RESTART = "Restart"
DESIRE_STOP = "Stop"
DESIRE_NONE = ""

ADDRESS_KEY = "Interface"
DESIRED_STATE_KEY = "DesiredState"

EX_PORTS_KEY = "ExternalPorts"
IN_PORTS_KEY = "InternalPorts"
ENV_KEY = "Environment"

PORT_SEPARATOR = "/"
ENV_SEPARATOR = "="

ALLOWED_PROTOCOLS = frozenset(("tcp", "udp"))

ALLOWED_DESIRED_STATES = frozenset((DESIRE_ACTIVE,
                                    DESIRE_RESTART,
                                    DESIRE_STOP,
                                    DESIRE_NONE))


ALLOWED_STATES = frozenset((ACTIVE,
                            AVAILABLE,
                            STOPPED,
                            RESTARTING))

from . import plugins
from . import interfaces


"""
quick helper functions
"""
from . import ADDRESS_KEY


def _check_common(field, interface, port_data):
    """

    :param field:
    :param interface:
    :param port_data:
    :return:
    """
    response = {}
    common = list(set(port_data[field]) &
                  set(interface[field]))
    if common:
        msg = "{} conflicts(s): {} in use on {}".format(field,
                                                        common,
                                                        interface[ADDRESS_KEY])
        response = {"errors": 1,
                    "first_error": msg}
    return response


def has_port_conflict(port_data,
                      existing):
    """

    :param port_data:
    :param existing:
    :return:
    """
    for interface in existing:
        common_tcp = _check_common("TCPPorts", interface, port_data)
        common_udp = _check_common("UDPPorts", interface, port_data)
        if common_tcp:
            return common_tcp
        elif common_udp:
            return common_udp
    return None


"""
decorators for controller specific functions
"""

from decorator import decorator
from . import ID_KEY, SERVICE_KEY


UNKNOWN_PLUGIN_ERROR = {
    "errors": 1,
    "first_error": "Cannot update non-existent plugin!"
}


@decorator
def expect_arg_type(func_, expected=None,
                    *args, **kwargs):
    """

    :param func_:
    :param expected:
    :param args:
    :param kwargs:
    :return:
    """
    assert isinstance(expected, tuple)
    for arg_idx in range(len(expected)):
        assert isinstance(args[arg_idx], expected[arg_idx])
    return func_(*args, **kwargs)


@decorator
def set_plugin_id(func_, *args, **kwargs):
    """

    :param func_:
    :param args:
    :param kwargs:
    :return:
    """
    from .plugins import find_plugin  # load time workaround- JIT import
    plugin = args[0]
    plugin_id = plugin.get(ID_KEY, False)
    service_name = plugin.get(SERVICE_KEY, False)
    if not plugin_id:
        found_plugin = find_plugin(service_name, SERVICE_KEY, args[-1])
        if found_plugin:
            plugin[ID_KEY] = found_plugin[-1][ID_KEY]
        else:
            return UNKNOWN_PLUGIN_ERROR
    return func_(*args, **kwargs)






"""
decorator objects for other functions in the main brain module
"""

from decorator import decorator
from sys import stderr

COMMAND_FIELD = "JobCommand"
INPUT_FIELD = "Inputs"
OPTIONAL_FIELD = "OptionalInputs"
VALUE_FIELD = "Value"


@decorator
def verify_jobs_args_length(func_, *args, **kwargs):
    """

    :param job:
    :param inputs:
    :param optional_inputs:
    :return:
    """
    job = args[0]
    inputs = args[1]
    optional_inputs = args[2]
    assert len(job[COMMAND_FIELD][INPUT_FIELD]) == len(inputs)
    if not optional_inputs:
        optional_inputs = tuple()
    assert len(job[COMMAND_FIELD][OPTIONAL_FIELD]) == len(optional_inputs)
    return func_(*args, **kwargs)


@decorator
def verify_jobs_args_is_tuple(func_, *args, **kwargs):
    """

    :param func_:
    :param args:
    :param kwargs:
    :return:
    """
    inputs = args[1]
    optional_inputs = args[2]
    if not isinstance(inputs, tuple) or not isinstance(optional_inputs, tuple):
        raise ValueError('Input must be a tuple')
    return func_(*args, **kwargs)


@decorator
def deprecated_function(func_, replacement="(see docs)",
                        *args, **kwargs):
    """
    decorator to annotate deprecated functions

    usage @decorator(replacement="brain.whatever.new_function")

    :param func_: <callable>
    :param replacement: <str>
    :param args:  positional arguments
    :param kwargs:
    :return: <func_'s return value>
    """
    msg = "{} is deprecated, use {}\n".format(func_.__name__,
                                              replacement)
    stderr.write(msg)
    return func_(*args, **kwargs)
